#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_main.h>

#include <db.h>
#include <gd.h>
#include <gdfontg.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define MAX_CSTRING             256
#define MAX_QSTRING             1024

struct rgb {
    int		rgb_r;
    int		rgb_g;
    int		rgb_b;
};

/* image type 0 = jpeg, 1 = png */

struct countent {
    char	*ce_key;
    char	*ce_font;
    int		ce_imgtype;
    int		ce_val;
    int 	ce_ignore;
    int		ce_width;
    long	ce_reset;
    int		ce_point;
    struct rgb	ce_bgcolor;
    struct rgb	ce_text;
};

/* these are the defaults if no directives are given */
static char	*counter_db = "/var/odometer/odometer.db";
static char	*font_dir = "/var/odometer/font/";
static char	*lockfile = NULL;

static int	lock( void );
static int	unlock( void );
static void	odometer_init( server_rec *s, pool *p );
static int	update_counter( struct countent * );
static void	rgb_color( const char *, struct rgb * );
static int	apmod_sink( void *, char *, int );
static int	display_counter( int, struct countent *, request_rec * );
const char *	set_counter_db( cmd_parms *, void *, char * );
const char *	set_font_dir( cmd_parms *, void *, char * );

    static void
odometer_init( server_rec *s, pool *p )
{
    char *prefix, *c;

    prefix = ap_pstrdup( p, counter_db );

    if (( c = strrchr( prefix, '/' )) == NULL ) {
	lockfile = ap_pstrcat( p, "./odometer.lock", NULL );
    } else {
	*c = '\0';
	lockfile = ap_psprintf( p, "%s/odometer.lock", prefix );
    }

    if ( unlink( lockfile ) == 0 ) {
	ap_log_error( APLOG_MARK, APLOG_INFO|APLOG_NOERRNO, s,
		"mod_ometer: unlinked %s.", lockfile );
    }
}

    static int
lock( void )
{
    int         tries = 10, fd = -1;

    while ( tries > 0 ) {
        if (( fd = open( lockfile, O_RDWR | O_CREAT | O_EXCL, 0 )) >= 0) {
            break;
        }
        sleep( 1 );
        tries--;
    }

    if ( fd < 0 ) {
        return( 0 );
    }

    (void)close( fd );

    return( 1 );
}

    static int
unlock( void )
{
    if ( unlink( lockfile ) < 0 ) {
        return( 0 );
    }

    return( 1 );
}

    static int
update_counter( struct countent *ce )
{
    int		ret;	
    DB		*dbp;
    DBT		key, value;

    memset( &key, 0, sizeof( key ));
    memset( &value, 0, sizeof( value ));
    key.data = ce->ce_key;
    key.size = strlen( ce->ce_key );

    if (( ret = db_create( &dbp, NULL, 0 )) != 0 ) {
	fprintf(stderr, "db_create: %s\n", db_strerror(ret));
	return( -1 );
    }

    if (( ret = dbp->open( dbp, counter_db, NULL,
	    DB_HASH, DB_CREATE, 0664)) != 0 ) {
	dbp->err( dbp, ret, "%s", counter_db );
	return( -1 );
    }

    if ( !lock() ) {
        fprintf( stderr, "Unable to lock databse.\n" );
        goto cleanup;
    }

    if (( ret = dbp->get( dbp, NULL, &key, &value, 0 )) == 0 ) {
	memcpy( &ce->ce_val, value.data, value.size);
	if ( !ce->ce_ignore && ce->ce_reset == -1 ) { 
	    ce->ce_val++;
	}
	if ( ce->ce_reset != -1 ) {
	    ce->ce_val = ce->ce_reset;
	}
    } else if ( ret == DB_NOTFOUND ) {
	ce->ce_val = 1;
    } else {
	dbp->err( dbp, ret, "DB->get");
	goto cleanup2;
    }

    value.data = &ce->ce_val;
    value.size = sizeof( ce->ce_val );

    if (( ret = dbp->put( dbp, NULL, &key, &value, 0 )) != 0 ) {
	dbp->err( dbp, ret, "DB->put" );
	goto cleanup2;
    }

    if (( ret = dbp->close( dbp, 0 )) != 0 ) {
        fprintf( stderr, "db: close failed\n" );
    }

    if ( !unlock() ) {
        perror( lockfile );
    }

    return( ce->ce_val );

cleanup2:
    if ( !unlock() ) {
        perror( lockfile );
    }
cleanup:
    if (( ret = dbp->close( dbp, 0 )) != 0 ) {
        fprintf( stderr, "db: close failed\n" );
    }
    return( -1 );

}

    static void
rgb_color( const char *hexnum, struct rgb *rgb )
{
    long	hexcolor;
    int		mask = 0xFF;

    hexcolor = strtol( hexnum, (char **)NULL, 16 );
    rgb->rgb_b = ( hexcolor & mask );
    rgb->rgb_g = (( hexcolor >> 8 ) & mask );
    rgb->rgb_r = (( hexcolor >> 16 ) & mask );

    return;
}



    static int
apmod_sink( void *r, char *buf, int len )
{
    /* instead of writing  on stdout, use apache's api */
    return( ap_rwrite( buf, len, r ));
}

    static int
display_counter( int cn, struct countent *ce, request_rec *r ) 
{
    gdSink		ap_sink;
    gdImagePtr		im;
    int			bgcolor, text, x, y, usepng = 0; 
    int			brect[ 8 ];
    FILE		*jpeg = NULL;
    char		*num = NULL, *odofile = NULL, *err = NULL;
    const char		*accept = NULL;

    num = ap_psprintf( r->pool, "%.*d", ce->ce_width, cn );
    ap_sink.context = r;
    ap_sink.sink = (void *)apmod_sink;

    if ( ce->ce_font == NULL ) {
	x = gdFontGiant->h + 2;
	y = gdFontGiant->w * ( strlen( num ) + 1 );

	if (( im = gdImageCreate( y, x)) == NULL ) {
	    return( -1 );
	}

	/* Background color (first allocated) */
	bgcolor = gdImageColorAllocate( im, ce->ce_bgcolor.rgb_r,
		ce->ce_bgcolor.rgb_g, ce->ce_bgcolor.rgb_b );
	text = gdImageColorAllocate( im, ce->ce_text.rgb_r,
		ce->ce_text.rgb_g, ce->ce_text.rgb_b );
	gdImageString( im, gdFontGiant,
		im->sx / 2 - ( strlen( num ) * gdFontGiant->w / 2 ),
		im->sy / 2 - gdFontGiant->h / 2, num, text );
    } else {

	/* obtain brect so that we can size the image */
	if (( err = gdImageStringFT( NULL, &brect[ 0 ], 0, ce->ce_font,
		ce->ce_point, 0., 0, 0, num )) != NULL ) {
	    ap_log_error( APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r->server,
		    "gdImageStringFT: %s", err );
	    return( -1 );
	}

	x = brect[ 2 ]-brect[ 6 ] + 10;
	y = brect[ 3 ]-brect[ 7 ] + 10;

	im = gdImageCreate( x,y );

	/* render the string, offset origin to center string
	   note that we use top-left coordinate for adjustment 
	   since gd origin is in top-left with y increasing downwards. */

	x = 3 - brect[ 6 ];
	y = 3 - brect[ 7 ];

	/* Background color (first allocated) */
	bgcolor = gdImageColorResolve( im, ce->ce_bgcolor.rgb_r,
		ce->ce_bgcolor.rgb_g, ce->ce_bgcolor.rgb_b );
	text = gdImageColorResolve( im, ce->ce_text.rgb_r,
			ce->ce_text.rgb_g, ce->ce_text.rgb_b );
	if (( err = gdImageStringFT( im, &brect[ 0 ], text, ce->ce_font, 
		ce->ce_point, 0.0, x, y, num )) != NULL ) {
	    ap_log_error( APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r->server,
		    "gdImageStringFT: %s", err );
	    gdImageDestroy( im );
	    return( -1 );
	}

    }

    r->content_type = "image/jpeg";
    if ( ce->ce_imgtype ) {
	accept = ap_table_get( r->headers_in, "Accept" );
	if ( strstr( (char *)accept, "image/png" )) {
	    r->content_type = "image/png";
	    usepng = 1;
	}
    }

    ap_table_set(r->headers_out, "Pragma", "no-cache");
    ap_table_set(r->headers_out, "Cache-Control", "no-cache");
    r->no_cache = 1;

    ap_hard_timeout( "mod_ometer", r );
    if ( usepng ) {
	ap_send_http_header( r );
	gdImagePngToSink( im, &ap_sink );
    } else {
	odofile = ap_psprintf( r->pool, "/tmp/odo.jpg.%i", (int)getpid());
	jpeg = ap_pfopen( r->pool, odofile, "wb+" );
	if ( unlink( odofile ) < 0 ) {
	    fprintf( stderr, "%s failed to unlink\n", odofile );
	}
	gdImageJpeg( im, jpeg, 95 );
	if ( fseek( jpeg, 0L, SEEK_SET ) != 0 ) {
	    perror( "lseek" );
	    ap_pfclose( r->pool, jpeg );
	    gdImageDestroy( im );
	    return( -1 );
	}
	ap_send_http_header( r );
	ap_send_fd( jpeg, r );
	ap_pfclose( r->pool, jpeg );
    }
    gdImageDestroy( im );
    ap_kill_timeout( r );

    return( 0 );
}

    static int 
odometer_handler(request_rec *r) 
{
    const char		*hostname = NULL, *ipaddr = NULL;
    struct countent	ce = {
				NULL, NULL, 0, 1, 0, 0, -1, 20,
				{ 0, 0, 0 },
				{ 255, 255, 255 },
			    }; 
    int 		random = 0, cn = 0;
    const char 		*data = r->args, *param;
    const char     	*re = NULL, *p = NULL, *ig = NULL, *ra = NULL;
    const char     	*wi = NULL, *fo = NULL, *bg = NULL, *tx= NULL;
    const char     	*pt = NULL, *it = NULL;

    hostname = ap_get_remote_host(
	    r->connection, r->per_dir_config, REMOTE_NAME);
    ipaddr = r->connection->remote_ip;
    if ( hostname == NULL || ipaddr == NULL ) {
	goto err;
    }

    if ( r->method_number == M_POST ) {
	goto err;
    }

    while ( data && *data && ( param = ap_getword( r->pool, &data, '&' ))) {
    
	if (( strlen( data ) > MAX_QSTRING ) || ( param == NULL )) {
	    goto err;
	}
        p = ap_getword( r->pool, &param, '=' );
        if ( p == NULL ) {
            goto err;
        }
        if ( strcasecmp( p, "link") == 0 ) {
            (const char *)ce.ce_key = param;
	    ap_unescape_url( ce.ce_key );
        } else if ( strcasecmp( p, "reset") == 0 ) {
            re = param;
	    ce.ce_reset = strtol( re, (char **)NULL, 10 );
        } else if ( strcasecmp( p, "width") == 0 ) {
            wi = param;
	    if (( ce.ce_width = strtol( wi, (char **)NULL, 10 )) > 10 ) {
		ce.ce_width = 10;
	    }
        } else if ( strcasecmp( p, "font") == 0 ) {
	    fo = param;
	    if ( strstr( fo, "../" ) == NULL ) { 
		ce.ce_font = ap_psprintf( r->pool, "%s%s.ttf", font_dir, fo );
		if ( access( ce.ce_font,  R_OK ) < 0 ) {
		    ce.ce_font = NULL;
		}
	    }
        } else if (( strcasecmp( p, "ignore") == 0 ) && ce.ce_ignore == 0 ) {
            ig = param;
	    if (( strcmp( ig, hostname ) == 0 ) ||
		    ( strcmp( ig, ipaddr ) == 0 )) {
		ce.ce_ignore = 1;
	    }
	} else if ( strcasecmp( p, "bgcolor") == 0 ) {
	    if (( bg = param) != NULL ) {
		rgb_color( bg, &ce.ce_bgcolor );
	    }
	} else if ( strcasecmp( p, "text") == 0 ) {
	    if (( tx = param) != NULL ) {
		rgb_color( tx, &ce.ce_text );
	    }
	} else if ( strcasecmp( p, "point") == 0 ) {
	    pt = param;
	    if (( ce.ce_point = strtol( pt, (char **)NULL, 10 )) > 128 ) {
		ce.ce_point = 20;
	    }
	} else if ( strcasecmp( p, "image") == 0 ) {
	    it = param;
	    if ( strcasecmp( it, "png" ) == 0 ) {
		ce.ce_imgtype = 1;
	    }
        } else if ( strcasecmp( p, "rand") == 0 ) {
            ra = param;
	    if ( strcasecmp( ra, "true" ) == 0 ) {
		random = 1;
		srand( (unsigned int)r->request_time ^ getpid(  )); 
		cn = rand( );
		/* skip all database stuff */
	    }
	}
    }

    if ( ce.ce_key == NULL && !random ) {
	goto err;
    }

    /* skip all database stuff if it's random */
    if ( !random ) {
	if (( cn = update_counter( &ce )) < 0 ) {
	    goto err;
	}
    }

    if ( display_counter( cn, &ce, r ) == 0 ) {
	return OK;
    }

err:
    ce.ce_font = NULL;
    ce.ce_imgtype = 1;
    rgb_color( "000000", &ce.ce_bgcolor );
    rgb_color( "FFFFFF", &ce.ce_text );
    if ( display_counter( 404, &ce, r ) == 0 ) {
	return OK;
    }
    ap_log_error( APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r->server,
	    "mod_ometer: unable to display 404" );
    return NOT_FOUND;
}

    const char *
set_counter_db( cmd_parms *parms, void *dummy, char *arg )
{
    counter_db = arg;
    return NULL;
}

    const char *
set_font_dir( cmd_parms *parms, void *dummy, char *arg )
{
    font_dir = arg;
    return NULL;
}

command_rec odometer_cmds[ ] =
{
	{ "OdometerDB", set_counter_db,
	NULL, RSRC_CONF, TAKE1,
	"the name of the counter database file"},

	{ "OdometerFontDir", set_font_dir,
	NULL, RSRC_CONF, TAKE1,
	"the name of the font directory"},

	{ NULL }
};

static handler_rec odometer_handlers [] =
{
    { "odometer-handler", odometer_handler },
    { NULL }
};


/* Tell Apache what phases of the transaction we handle */
module MODULE_VAR_EXPORT odometer_module =
{
    STANDARD_MODULE_STUFF,
    odometer_init,      /* module initializer                 */
    NULL,		/* per-directory config creator       */
    NULL,  		/* dir config merger                  */
    NULL,               /* server config creator              */
    NULL,               /* server config merger               */
    odometer_cmds,      /* command table                      */
    odometer_handlers,  /* [7]  content handlers              */
    NULL,               /* [2]  URI-to-filename translation   */
    NULL,               /* [5]  check/validate user_id        */
    NULL,               /* [6]  check user_id is valid *here* */
    NULL,               /* [4]  check access by host address  */
    NULL,               /* [7]  MIME type checker/setter      */
    NULL,               /* [8]  fixups                        */
    NULL,               /* [9]  logger                        */
    NULL,               /* [3]  header parser                 */
    NULL,               /* process initialization             */
    NULL,               /* process exit/cleanup               */
    NULL                /* [1]  post read_request handling    */
};
